% MATLAB code in support of menke_gls_paper
% Example 2:  GLS implementation of a very
% simple linearized inversion, Bill Menke, Feb 2014
clear all

% globals necessary to communicate constants
% to functions that compute g(m) and its derivative
global c1 c2 x

% globals necessary to communicate inverse problem
% to function called by biconjugate gradient solver
global G H vard varh

% simplified equation
% d(i) = c2 * exp( m(1) + c1*sin(x(i)))
% that is, with (d,x) paired vectors and c1, c2 scalars

% make a vector of x's
N=100;
M = 1;

% allocate matrices & vectors
dtrue = zeros(N,1);
dobs = zeros(N,1);
dpre = zeros(N,1);
d_0 = zeros(N,1);
dd = zeros(N,1);
mtrue = zeros(M,1);
mpri = zeros(M,1);
m_p = zeros(M,1);
m_0 = zeros(M,1);
mest = zeros(M,1);
dm = zeros(M,1);
hpri = zeros(M,1);
dh = zeros(M,1);
G = zeros(N,M);
H = zeros(M,M);

% auxillary variable x
Dx = pi/8;
x = Dx*[0:N-1]';

% define scalars as unity
c1 = 1.0;
c2 = 1.0;

% case 1
mtrue(1) = 2.0;
% prior information
H = eye(M,M);
mpri(1) = 2.2;
hpri = mpri; % prior value of parameret a
sigmah = 0.5; % prior standard deviation of parameter a
varh = (sigmah^2)*ones(M,1);

dtrue = g(mtrue); % true d
sigmad = 0.5; % noise level for d
vard = (sigmad^2)*ones(N,1);
dobs = dtrue + random('Normal',0,sigmad,N,1); % observed d is noisy
% top plot
% black curve: true data, dtrue
% black circles: observed (noisy) data, dobs
% green curve: predicted data based on wildly wrong initial guess, m0_guess
% red curve: predicted data based on least-squared estimate of mest
figure(1);
clf;
set(gca, 'LineWidth', 2 );
axis( [0, 40, 0, 50] );
hold on;
plot( x, dtrue, 'k-', 'LineWidth', 3 );
plot( x, dobs, 'ko', 'LineWidth', 2 );
xlabel('x');
ylabel('d');
title('data: true (black curve), data (black circles), starting (green), final (red)');

% guess for m and corresponding data
m_0 = 2.5;
d_0 = g(m_0);
plot( x, d_0, 'g-', 'LineWidth', 2 );

% solution by iterating linearized problem
m_p = m_0;
for itt=[1:10]
    d_p = g(m_p);
    dd = dobs - d_p;
    dh = hpri - H*m_p;
    G = dgdm(m_p); % matrix of derivatives dg/dm 
    TOL=1e-5;
    MAXI=4*N;
    dm = bicg(@glsfcn,G'*(dd./vard)+H'*(dh./varh),TOL,MAXI);
    m_p = m_p+dm;
    if( ((dm'*dm)/(m_p'*m_p)) <= TOL )
        break;
    end
end
mest = m_p;

dpre = g(mest);
plot( x, dpre, 'r-', 'LineWidth', 2 );
fprintf('case 1: mtrue %f, hpri %f, sigmah %s, m_0 %f, mest %f\n', mtrue(1), hpri(1), sigmah, m_0, mest);

